% Matlab code example for Algorithm 1 presented in "Sparse Fast DCT for 
% Vectors with One-block Support" by Sina Bittens and Gerlind Plonka.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [ y, supp_struct, mu, m ] = DFT_reflected(yhat,epsilon,varargin)
%DFT_REFLECTED Reconstructs the vector y with reflected block support from 
%its Fourier transform yhat.
% epsilon is the threshold value for cutting off noise. An a priori lower
% bound on the block length m of y is known such that 2^(b-1)<=m can be
% given as vargin.

% The length of yhat has to be a power of 2. Write it as 2^J and set
% N=2^(J-1).
N=length(yhat)/2;
if abs(log2(N)-round(log2(N)))<eps 
    J=log2(N)+1;
    % Check whether a lower bound on the block length of y is known. 
    if nargin==3
        b=varargin{:};
        if b>J-1
            error('2^b can be at most length(y_{(0)}.')
        end
    else
        % If no bound is known, start with a vector of length 2^0=1.
        b=0;
    end
    % Compute the smallest periodized vector. If b=0, then y=yhat(1).
    y=ifft(yhat((2^(J-b)*(0:2^b-1))+1));
    % If b==0, then y has a one-block support. Otherwise detect its support
    % structure.
    if b==0
        supp_struct=1;
        mu=0;
        m=1;
    else
        [supp_struct, mu, m]=find_supp_0(y,epsilon); 
    end
    % In each step recover y^(j+1) from y^(j) and additional entries of
    % yhat using the appropriate method for the given support structure.
    for j=b:J-1
        %% Recovery Step for a Periodization with One-block Support
        if supp_struct==1
            % If m>2^(j-1), y^(j) may not have a uniquely defined first
            % support index and y^(j) does not have a short support we can
            % benefit from. Hence, recover y^(j+1) via an IFFT of length
            % 2^j.
            if m>2^(j-1)
                data=yhat(2^(J-j-1)*(2*(0:2^j-1)+1)+1);
                a=exp(pi*1i*(0:2^j-1)/2^j).'.*ifft(data);    
                % Since y is a non-negative real vector, we can disregard
                % the imaginary part. Further, we apply a thresholding
                % procedure in order to be able to better determine the
                % support structure of y^(j+1).
                y_0=real(0.5*(y+a));
                aux=find(y_0<=epsilon);
                y_0(aux)=zeros(size(aux));
                % Use that all periodizations of y are symmetric.
                y_1=y_0(end:-1:1);
                y=[y_0;y_1]; 
                % Find the support structure of y^(j+1).
                [supp_struct, mu, m]=find_supp_1(y,j,J,mu,m,epsilon);
            % If m<=2^(j-1), the first support index of y^(j) is uniquely
            % determined and we can benefit from the fact that the support 
            % of y^(j) is short.
            else
                L=ceil(log2(m));
                % Restrict to 2^L equations and variables.
                data=yhat(2^(J-L)*(0:2^L-1)+2^(J-j-1)+1);
                z=y(mod(mu+(0:(2^L-1)),2^j)+1);
                % If mu>2^j-2^L, the modulo operation in the first matrix
                % has to be taken into account.
                if mu>2^j-2^L
                   a=[exp(pi*1i*(mu+(0:2^j-mu-1))/2^j) -exp(pi*1i*(mu+(2^j-mu:2^L-1))/2^j)].'.*ifft(exp(2*pi*1i*mu*(0:2^L-1)/2^L).'.*data);
                else
                   a=exp(pi*1i*(mu+(0:2^L-1))/2^j).'.*ifft(exp(2*pi*1i*mu*(0:2^L-1)/2^L).'.*data);
                end 
                % Since y is a non-negative real vector, we can disregard
                % the imaginary part. Further, we apply a thresholding
                % procedure in order to be able to better determine the
                % support structure of y^(j+1).
                z=real(0.5*(z+a));
                aux=find(z<=epsilon);
                z(aux)=zeros(size(aux));
                y_0=zeros(2^j,1);
                y_1=zeros(2^j,1);
                y_0(mod(mu+(0:(2^L-1)),2^j)+1)=z(1:2^L);
                % Use that all periodizations of y are symmetric.
                y_1(2^j-1-mod(mu+(0:2^L-1),2^j)+1)=z(1:2^L);
                y=[y_0;y_1];
                % Find the support structure of y^(j+1).
                [supp_struct, mu, m]=find_supp_2(y,j,J,mu,m,epsilon);
            end
        %% Recovery Step for a Periodization with Two-Block Support
        elseif supp_struct==2
            % If y^(j) has a two-block support, then y^(j+1) has a 
            % two-block support as well and we only need to find the first 
            % support index of the first block of y^(j) in y^(j+1).
            % Find a non-zero oddly indexed entry of fft(y^(j+1)) by only
            % considering O(m) entries.
            [~,k]=max(abs(yhat(2^(J-j-1)*(2*(0:2*m-1)+1)+1)));
            k=k-1;
            yhat_odd=yhat(2^(J-j-1)*(2*k+1)+1);            
            % Compute the 2k+1-st entry of the DFT of the vector u that has
            % y^(j) as its periodization and the same first support index 
            % as y^(j).
            uhat=sum(y((mu:mu+m-1)+1).*(exp(-pi*1i*(2*k+1)/2^j*(mu:mu+m-1))).')+sum(y(2^j-1-(mu:mu+m-1)+1).*(exp(pi*1i*(2*k+1)/2^j*(mu+1:mu+m))).');
            % If uhat and yhat_odd are equal, then the first support index
            % lambda of the first block of y^(j) in y^(j+1) is mu, so y^(j)
            % and y^(j+1) have the same first support index.
            if abs(uhat-yhat_odd)<abs(uhat+yhat_odd)
                lambda=mu;
            % Otherwise the first support index of that block in y^(j+1) is
            % 2^j+mu.
            else
                lambda=2^j+mu;
            end
            y_0=zeros(2^(j+1),1);
            y_0((lambda:lambda+m-1)+1)=y(mod((lambda:lambda+m-1),2^j)+1);
            % Use that all periodizations of y are symmetric.
            y_0((2^(j+1)-m-lambda:2^(j+1)-1-lambda)+1)=y(mod((2^(j+1)-m-lambda:2^(j+1)-1-lambda),2^j)+1);
            y=y_0;
            % Find the support structure of y^(j).
            [supp_struct, mu, m]=find_supp_3(j,mu,m,lambda);
        end
    end  
else
    disp('The length of dct_x has to be a power of 2.')
end

end

