% Required for the Matlab code example for Algorithm 1 presented in "Sparse 
% Fast DCT for Vectors with One-block Support" by Sina Bittens and Gerlind
% Plonka.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [  ] = TEST_DFT_reflected( tests,m,N,epsilon,variant )
%TEST_DFT_REFLECTED Computes the average runtime and reconstruction error
%of DFT_reflected, MATLAB's ifft and algorithmshort, the general M-sparse 
%IDFT procedure presented as Algorithm 2.3 in "Deterministic sparse FFT for
%M-sparse vectors" by Gerlind Plonka, Katrin Wannenwetsch, Annie Cuyt and
%Wen-shin Lee over test tries for an input vector of length N with 
%reflected block support.

% If variant==0, the average runtimes of DFT_reflected, algorithmshort and 
% ifft are computed for randomly generated noiseless data. If variant==1, 
% the average reconstruction errors for randomly generated noisy input 
% vectors (uniform noise with SNR values [0 10 ... 50]) are computed for 
% DFT_reflected and ifft, as algorithmshort is not stable enough for noisy 
% input data.

% epsilon is the threshold for cutting off noise required by DFT_reflected 
% in order to find the support structure of the reconstructed vectors.

% For DFT_reflected and algorithmshort the length of the input vector has
% to be a power of 2.
if abs(log2(N)-round(log2(N)))<eps
    J=log2(N)+1;
    M=length(m);
    % Initialize the matrix for the average runtimes.
    time=zeros(M,5);
    time(:,1)=m;
    % Define the SNR values for noisy signals.
    SNR=zeros(6,1);
    for l=1:length(SNR)
        SNR(l)=(l-1)*10;
    end
    % Initialize the matrix of reconstruction errors for exact data.
    error_ex=zeros(M,3);    
    for j=1:M
        % Initialize the matrix for the average runtimes for noisy data.
        time_noise=zeros(length(SNR),4);
        time_noise(:,1)=SNR;
        % Initialize the matrix for the average noisy reconstruction 
        % errors.
        error=zeros(length(SNR),3);
        error(:,1)=SNR;
        % Initialize the matrix for the rate of correct recovery.
        correct_recov=zeros(length(SNR),4);
        correct_recov(:,1)=SNR;
        % Define the cutoff thresholds for noisy data, optimized for
        % N=2^20.
        epsilon_noise(6)=0.02;
        epsilon_noise(5)=0.05;
        epsilon_noise(4)=0.19;
        epsilon_noise(3)=0.4;
        epsilon_noise(2)=1.2;
        epsilon_noise(1)=1.7;
        for t=1:tests
            if mod(t,10)==1
                t
            end
            % For each test create a random first support index mu for the
            % support block in the first half of y, which has to be a
            % vector of length 2N with reflected block support.
            mu=randi([0 N-1],1);
            y=zeros(2*N,1);
            % Create the m(j) support indices.
            supp=[0:mu+m(j)-1-N,mu:min(mu+m(j)-1,N-1)];             
            % Generate a random number between 0 and floor((m(j)-2)/2) and
            % create that many zeros inside the support, but not at the 
            % first or last support index.
            suppzeros=randi(floor((m(j)-2)/2)+1)-1;
            % Select suppzeros elements randomly from 2,...,m-1 and remove
            % the respective entries from supp.            
            supp(randperm(m(j)-2,suppzeros)+1)=[];
            % Create the m(j)-suppzeros random non-negative entries of the 
            % support.
            y(supp+1)=rand(m(j)-suppzeros,1)*10;
            % The second half of y is already determined by the first one,
            % as y is symmetric.
            y(2*N-1-supp(end:-1:1)+1)=y(supp(end:-1:1)+1);            
            % Create the input data.
            yhat=fft(y);
            
            %% Standard IFFT Algorithm
            % Measure the runtime of ifft.
            tic;
            S_ifft=ifft(yhat);
            time(j,4)=time(j,4)+toc;
            
            %% Numerical Experiments for Exact Input Data
            if variant==0
                %% DFT_reflected
                % Measure the runtime of DFT_reflected.
                tic;
                [S_reflected,~,~,~] =DFT_reflected(yhat,epsilon);
                time(j,2)=time(j,2)+toc;
                
                %% algorithmshort
                % Measure the runtime of the general M-sparse IFFT
                % procedure algorithmshort. Since the algorithm is only
                % stable for small sparsities 2*m(j), we only measure the
                % runtime for block lengths at most 30.
                if m(j)<=30
                    tic;
                    [yrec, activeind, ~] = algorithmshort(yhat);
                    time(j,3)=time(j,3)+toc;
                    % algorithmshort only returns length(activeind) entries
                    % of y in yrec for the indices that were found to be
                    % active by the procedure. Hence, we have to insert
                    % yrec correctly into a 2*N-length vector in order to
                    % compare it to the noiseless vector y.
                    S_msparse=zeros(2*N,1);
                    S_msparse(activeind+1)=yrec;
                    % Compute the reconstruction error for exact data, 
                    % given by the norm of the difference between the 
                    % the original vector y and the returned vector
                    % S_msparse.
                    error_ex(j,2)=error_ex(j,2)+norm(S_msparse-y)/(2*N);
                end
                % Compute the reconstruction errors for exact data for the 
                % two remaining methods, given by the norm of the 
                % difference between the original vector y and the returned
                % vectors S_reflected and S_ifft.
                error_ex(j,1)=error_ex(j,1)+norm(S_reflected-y)/(2*N);
                error_ex(j,3)=error_ex(j,3)+norm(S_ifft-y)/(2*N);  
            
            %% Numerical Experiments for Noisy Input Data
            elseif variant==1
                for n=1:length(SNR)
                    % Create random uniformly distributed noise for the 
                    % input vector yhat.
                    noise=rand(2*N,1)-0.5+1i*(rand(2*N,1)-0.5);          
                    % Scale the noise according to the current SNR value
                    % SNR(n).
                    noise=norm(yhat)*10^(-SNR(n)/20)/norm(noise)*noise; 
                    % Create noisy input data by addition.
                    yhat_noise=yhat+noise;
                    
                    %% Standard IFFT for Noisy Input Data
                    % Measure the runtime of ifft for noisy input data.
                    tic;
                    S_ifft_noise=ifft(yhat_noise);
                    time_noise(n,4)=time_noise(n,4)+toc;
                    
                    %% DFT_reflected for Noisy Input Data
                    % Measure the runtime of DFT_reflected for noisy input
                    % data.
                    tic;
                    [S_reflected_noise, ~, mu_ref_n, m_ref_n]=DFT_reflected(yhat_noise,epsilon_noise(n));
                    time_noise(n,2)=time_noise(n,2)+toc;
                    % Check whether the support of y was found correctly.
                    % First examine whether the support of y is contained
                    % in the returned support of S_reflected_noise, and, if
                    % necessary, increase the counter correct_recov(n,2).                    
                    if mu_ref_n+m_ref_n<=2^J-1
                        if (mu>=mu_ref_n && mu+m(j)-1<=mu_ref_n+m_ref_n-1) 
                            correct_recov(n,2)=correct_recov(n,2)+1;
                            % We also introduce a second counter
                            % correct_recov(n,3) which is only increased if
                            % the returned block length of
                            % S_reflected_noise is at most three times the
                            % length of the block length of y.
                            if (m_ref_n <= 3* m(j) && m_ref_n>=m(j))
                                correct_recov(n,3)=correct_recov(n,3)+1;
                            end
                        end
                    else
                        % If 0, 2^J-1 are contained in the support of
                        % S_reflected_noise, the block has to be considered
                        % modulo 2^J.
                        if ((mu>=mu_ref_n || mu<=mu_ref_n+m_ref_n-1-2^J) && (mu+m(j)-1>=mu_ref_n || mu+m(j)-1<=mu_ref_n+m_ref_n-1-2^J))
                            correct_recov(n,2)=correct_recov(n,2)+1;
                            if  (m_ref_n <= 3* m(j) && m_ref_n>=m(j))
                                correct_recov(n,3)=correct_recov(n,3)+1;
                            end
                        end
                    end
                    % Compute the reconstruction errors for noisy data for 
                    % the two methods, given by the norm of the difference 
                    % between the original vector y and the returned
                    % vectors S_ifft_noise and S_reflected_noise.
                    error(n,2)=error(n,2)+norm(S_reflected_noise-y)/(2*N);
                    error(n,3)=error(n,3)+norm(S_ifft_noise-y)/(2*N);
                end
            end 
        end
        % Compute the average runtimes and the average reconstruction 
        % errors.
        time(j,1)=m(j);
        time(j,5)=N;
        time_noise(:,2:4)=time_noise(:,2:4)/tests
        error(:,2:3)=error(:,2:3)/tests
        correct_recov(:,2:3)=correct_recov(:,2:3)/tests
        % Save the matrix of reconstruction error values in a separate file
        % for each block length m(j).
        if variant ==1
            filename = ['error_DFT_refl' num2str(m(j)) '.dat'];
            dlmwrite(filename, error(:,:), 'delimiter', '\t');
            % Save the matrix of correct recovery rates in a file.
            filename = ['correctrecov' num2str(m(j)) '.dat'];
            dlmwrite(filename,correct_recov,'delimiter','\t');
            % Save the matrix of noisy runtime values in a file.
            filename = ['runtimeDFT_refl_noise_' num2str(m(j)) '.dat'];
            dlmwrite(filename,time_noise,'delimiter','\t');
        end
    end
    time(:,2:4)=time(:,2:4)/tests
    error_ex=error_ex/tests
    if variant ==0
        % Save the matrix of runtime values in a file.
        dlmwrite('runtimeDFT_refl.dat',time(:,1:5),'delimiter','\t');
        % Save the matrix of exact errors in a file.
        dlmwrite('error_exact.dat',error_ex,'delimiter','\t');
    end
end

end

    


