import numpy as np
import math
from scipy.fftpack import ifft


class StrucVec:
    """A class for vectors with one- and two-block support."""
    def __init__(self, data, struct, mu, m):
        self.structure = struct
        self.vec = data
        self.supp_ind = mu
        self.supp_len = m

    def update_supp(self, new_struct, mu, m):
        self.structure = new_struct
        if self.structure == 'one' or self.structure == 'two':
            self.supp_ind = mu
            self.supp_len = m
        elif self.structure == 'two_diff':
            self.supp_ind = (mu[0], mu[1])
            self.supp_len = (m[0], m[1])


def dft_twoblock(yhat, epsilon, b=0):
    """
    Reconstructs the vector y with reflected block support from its Fourier transform yhat using Algorithm 1 in "Sparse
    fast DCT for vectors with one-block support" by Sina Bittens and Gerlind Plonka.
    :param yhat: Fourier transform of the vector y with reflected block support we aim to recover
    :param epsilon: noise threshold
    :param b: lower bound on the block length m of y = (x, (Jx)^T)^T: 2^(b-1) <= m
    :return x:
    :return supp_struct: support structure of y
             'one' if y has a one-block support
             'two' if y has a two-block support
             'two_diff' if y has a two-block support with possibly different lengths, only possible for j = J-1
    :return mu: first support index of y
    :return m: support length of y
    """

    yhat = yhat.ravel()
    N = round(yhat.shape[0] / 2)
    # The length 2N of y has to be a power of 2. Write it as 2^J.
    if abs(math.log(N, 2)-round(math.log(N, 2))) < np.spacing(1):
        J = round(math.log(N, 2)) + 1
        # The optional lower bound on the block length of y can be at most
        if b > J:
            print('Error: b = {} > J = {}.'.format(b, J))
            return
        # Compute the smallest periodized vector. If b == 0, then y = yhat[0].
        aux = ifft(yhat[2**(J - b) * np.arange(0, 2**b)])
        # If b == 0, then y has a one-block. Otherwise detect its support structure.
        if b == 0:
            y = StrucVec(aux, 'one', 0, 1)
        else:
            y = find_supp_0(aux, epsilon)

        # In each step recover y^(j+1) from y^(j) and additional entries of yhat using the appropriate method for the
        # given supprt structure.
        for j in range(b, J):
            # Recovery step for a periodization y^(j) with one-block support.
            if y.structure == 'one':
                # If m > 2^(j-1), y^(j) may not have a uniquely defined first support index and y^(j) does not have a
                # short support we can benefit from. Hence, recover y^(j+1) via an IFFT of length 2^j.
                if y.supp_len > 2 ** (j - 1):
                    data = yhat[2 ** (J - j - 1) * (2 * np.arange(0, 2 ** j) + 1)]
                    a = np.exp(np.pi * 1j * np.arange(0, 2 ** j) / 2 ** j) * ifft(data)
                    # Since y is a non-negative real vector, we can disregard the imaginary part. Further, we apply a
                    # thresholding procedure in order to be able to better determine the support structure of y^(j+1).
                    y_0 = np.real(0.5 * (y.vec + a))
                    aux = np.nonzero(y_0 <= epsilon)[0]
                    y_0[aux] = np.zeros(aux.shape)
                    # Use that all periodizations of y are symmetric.
                    y_1 = np.flip(y_0, 0)
                    y.vec = np.concatenate((y_0, y_1))
                    # Find the support structure of y^(j+1).
                    y = find_supp_1(y, j, J, epsilon)

                # If m <= 2**j, then the first support index of y^(j) is uniquely determined and we can benefit from
                # the fact that the support of y^(j) is short.
                else:
                    L = math.ceil(math.log(y.supp_len, 2))
                    # Restrict to 2^L equations and variables.
                    data = yhat[2 ** (J - L) * np.arange(0, 2 ** L) + 2 ** (J - j - 1)]
                    z = y.vec[(y.supp_ind + np.arange(0, 2 ** L)) % 2 ** j]
                    if y.supp_ind > 2 ** j - 2 ** L:
                        a = np.concatenate((np.exp(np.pi * 1j
                                                   * (y.supp_ind + np.arange(0, 2 ** j - y.supp_ind)) / 2 ** j),
                                            -np.exp(np.pi * 1j * (y.supp_ind
                                                                  + np.arange(2 ** j - y.supp_ind, 2 ** L)) / 2 ** j))) \
                            * ifft(np.exp(2 * np.pi * 1j * y.supp_ind * np.arange(0, 2 ** L) / 2 ** L) * data)
                    else:
                        a = np.exp(np.pi * 1j * (y.supp_ind + np.arange(0, 2 ** L)) / 2 ** j) \
                            * ifft(np.exp(2 * np.pi * 1j * y.supp_ind * np.arange(0, 2 ** L) / 2 ** L) * data)
                    # Since y is a non-negative real vector, we can disregard its imaginary part. Further, we apply a
                    # thresholding procedure in order to be able to better determine the support structure of y^(j+1).
                    z = np.real(0.5 * (z + a))
                    aux = np.nonzero(z <= epsilon)[0]
                    z[aux] = np.zeros(aux.shape)
                    y_0 = np.zeros(2 ** j)
                    y_1 = np.zeros(2 ** j)
                    y_0[(y.supp_ind + np.arange(0, 2 ** L)) % 2 ** j] = z[0: 2 ** L]
                    # Use that all periodizations of y are symmetric.
                    y_1[2 ** j - 1 - (y.supp_ind + np.arange(0, 2 ** L)) % 2 ** j] = z[0: 2 ** L]
                    y.vec = np.concatenate((y_0, y_1))
                    # Find the support structure of y^(j+1).
                    y = find_supp_2(y, j, J, epsilon)

            # Recovery step for a periodization y^(j+1) with two-block support.
            elif y.structure == 'two':
                # If y^(j) has a two-block support, then y^(j+1) has a two-block support as well and we only need to
                # find the first support index of the first block of y^(j) in y^(j+1).
                # Find an oddly indexed non-zero entry of fft(y^(j+1)) by only considering O(m) entries.
                k = np.argmax(abs(yhat[2 ** (J - j - 1) * (2 * np.arange(0, 2 * y.supp_len) + 1)]))
                yhat_odd = yhat[2 ** (J - j - 1) * (2 * k + 1)]
                # Compute the 2k+1-st entry of the DFT of the vector u that has y^(j) as its periodization and the same
                # first support index as y^(j).
                uhat = np.sum(y.vec[y.supp_ind: y.supp_ind + y.supp_len]
                              * (np.exp(-np.pi * 1j * (2 * k + 1) / 2 ** j
                                        * np.arange(y.supp_ind, y.supp_ind + y.supp_len)))) \
                       + np.sum(y.vec[2 ** j - 1 - np.arange(y.supp_ind, y.supp_ind + y.supp_len)]
                                * (np.exp(np.pi * 1j * (2 * k + 1) / 2 ** j
                                          * np.arange(y.supp_ind, y.supp_ind + y.supp_len))))
                # If uhat and yhat_odd are equal, then the first support index l of the first block of y^(j) in y^(j+1)
                # is mu, so y^(j) and y^(j+1) have the same first support index.
                if abs(uhat - yhat_odd) < abs(uhat + yhat_odd):
                    l = y.supp_ind
                # Otherwise the first support index of that block in y^(j+1) is 2^j+mu.
                else:
                    l = 2 ** j + y.supp_ind
                y_0 = np.zeros(2 ** (j + 1))
                y_0[l: l + y.supp_len] = y.vec[np.arange(l, l + y.supp_len) % 2 ** j]
                # Use that all periodizations of y are symmetric.
                y_0[2 ** (j + 1) - y.supp_len - l: 2 ** (j + 1) - l] \
                    = y.vec[np.arange(2 ** (j + 1) - y.supp_len - l, 2 ** (j + 1) - l) % 2 ** j]
                y.vec = y_0
                # Find the support structure of y^(j).
                y = find_supp_3(y, j, l)

        return y
    else:
        print('The length of y has to be a power of 2.')


def find_supp_0(y, epsilon):
    """
    Detects the support structure of y if nothing about its support is known.
    :param y: periodization of a vector with reflected block support
    :param epsilon: noise threshold
    :return: support structure:
             'one' if y has a one-block support
             'two' if y has a two-block support
    :return: mu: first support index
    :return: m: support length
    """
    y = y.ravel()
    N = y.shape[0]
    # Since y is symmetric, it suffices to examine the first half of it.
    y_0 = y[0: round(N / 2)]
    T = np.nonzero(y_0 > epsilon)[0]

    if T[0] == 0 and T[-1] != round(N / 2) - 1:
        # y has a one-block support of length m < N.
        supp_struct = 'one'
        mu = N - 1 - T[-1]
        m = 2 * (T[-1] + 1)
    elif T[-1] == round(N / 2) - 1 and T[0] != 0:
        # y has a one-block support of length m<N.
        supp_struct = 'one'
        mu = T[0]
        m = 2 * (T[-1] - T[0] + 1)
    elif T[-1] == round(N / 2) - 1 and T[0] == 0:
        # y has a one-block support of length m = N.
        supp_struct = 'one'
        mu = 0
        m = N
    else:
        # y has a two-block support of block length m < N".
        supp_struct = 'two'
        mu = T[0]
        m = T[-1] - T[0] + 1
    aux = StrucVec(y, supp_struct, mu, m)
    return aux


def find_supp_1(y, j, J, epsilon):
    """
    Finds the support structure of y = y^(j+1) if y^(j) has a one-block support of length m > 2^(j-1).
    :param y:
    :param j:
    :param J: full vector has length 2^J
    :param epsilon: noise threshold
    :return supp_struct: support structure of y
             'one' if y has a one-block support
             'two' if y has a two-block support
             'two_diff' if y has a two-block support with possibly different lengths, only possible for j = J-1
    :return mu_1: first support index of the first block of y^(j+1)
    :return m_1: length of the first block of y^(j+1)
    :return mu_2: first support index of the second block of y^(j+1)
    :return m_2: length of the second block of y^(j+1)
    """

    if j < J - 1:
        # If 2^(j-1)-1, 2^(j-1) are contained in S^(j) and supp((y^(j)) is not the whole interval 0,...,2^j-1, then
        # y has a two-block support.
        if (y.supp_ind + y.supp_len - 1 <= 2**j - 1 and ((2**(j-1) - 1 >= y.supp_ind) and
                                                         (2**(j-1) - 1 <= y.supp_ind + y.supp_len - 1)
                                                         and (2**(j-1) >= y.supp_ind)
                                                         and (2**(j-1) <= y.supp_ind + y.supp_len - 1)
                                                         and y.supp_len != 2**j)) \
                or (y.supp_ind + y.supp_len - 1 > 2**j - 1
                    and ((2**(j-1) - 1 >= y.supp_ind or 2**(j-1) - 1 <= (y.supp_ind + y.supp_len - 1) % 2**j)
                         and (2**(j-1) >= y.supp_ind or 2**(j-1) <= (y.supp_ind + y.supp_len - 1) % 2**j)
                         and y.supp_len != 2**j)):
            y_0 = y.vec[0: 2**j]
            # Find the first support index of the first block.
            T = np.nonzero(y_0 > epsilon)[0]
            y.update_supp('two', T[0], T[-1] - T[0] + 1)
            return y
        # Otherwise y has a one-block support.
        else:
            y_0 = y.vec[0: 2**j]
            # Find the significant entries in the first half of y.
            T_0 = np.nonzero(y_0 > epsilon)[0]
            K = T_0.shape[0]
            if K == 0:
                # If there are none, then the noisy data has cancelled out the significant entries. Set the first
                # support index to 0 and the length to 2^(j+1).
                y.update_supp('one', 0, 2**(j+1))
                return y
            else:
                # Otherwise detect whether the support block is centered around 0 and 2^(j+1)-1 or around 2^j-1 and 2^j.
                T = np.concatenate((T_0, 2**(j+1) - 1 - np.flip(T_0, 0)))
                e_0 = T[K] - T[K - 1]
                e_1 = (T[0] - T[2*K - 1]) % 2**(j+1)
                # If more than half of the entries of y are significant, the first support index is not uniquely defined
                # anymore. We set the first support index to 0 and the block length to 2^(j+1).
                # if T[2 * K - 1] - T[0] + 1 > 2**j:
                if 2 * K > 2 ** j:
                    y.update_supp('one', 0, 2**(j+1))
                    return y
                else:
                    if e_0 < e_1:
                        # If the distance between the last significant entry in the first half of y and the first
                        # significant entry in the second half is greater than the periodic distance between the first
                        # significant entry in the first half and the last significant entry in the second half, the
                        # support is centered around 2^j-1 and 2^j.
                        y.update_supp('one', T[0], T[2*K - 1] - T[0] + 1)
                    elif e_1 < e_0:
                        # If e_1 < e_0, the support is centered around 0 and 2^(j+1).
                        y.update_supp('one', T[K], 2**(j+1) - T[K] + T[K - 1] + 1)
                    else:
                        # If the distances are the same, both y^(j) and y^(j+1) have full support and we set the first
                        # support index to 0.
                        y.update_supp('one', 0, 2**(j+1))
                    return y

    else:
        # If j = J-1, we have the same cases as before, but it might also happen that y has a two-block support with
        # two blocks of different lengths.
        # If 2^(j-1)-1, 2^(j-1) are contained in S^(j) and the support of y^(j) is not the whole interval 0,...,2^j-1,
        # then y^(j+1) has a two-block support.
        # 2^(j-1)-1, 2^(j-1) in S^(j).
        if (y.supp_ind + y.supp_len - 1 <= 2**j - 1 and ((2**(j - 1) - 1 >= y.supp_ind)
                                                         and (2**(j - 1) - 1 <= y.supp_ind + y.supp_len - 1)
                                                         and (2**(j - 1) >= y.supp_ind)
            and (2**(j - 1) <= y.supp_ind + y.supp_len - 1) and y.supp_len != 2**j)) \
                or (y.supp_ind + y.supp_len - 1 > 2**j - 1 and ((2**(j - 1) - 1 >= y.supp_ind
                                                                 or 2**(j - 1) - 1
                                                                 <= (y.supp_ind + y.supp_len - 1) % 2**j)
                                                                and (2**(j - 1) >= y.supp_ind
                                                                     or 2**(j - 1)
                                                                     <= (y.supp_ind + y.supp_len - 1) % 2**j)
                                                                and y.supp_len != 2**j)):
            y_0 = y[0: 2**j]
            # Find the first support index of the first block.
            T = np.nonzero(y_0 > epsilon)[0]
            y.update_supp('two', T[0], T[-1] - T[0] + 1)
            return y
        # If 0, 2^j-1 are contained in S^(j), then y has a two-block support with possibly different block lengths, and
        # one of the blocks might be empty.
        elif (y.supp_ind + y.supp_len - 1 <= 2**j - 1 and ((0 >= y.supp_ind) and (0 <= y.supp_ind + y.supp_len - 1)
                                                           and (2**j - 1 >= y.supp_ind)
                                                           and (2**j - 1 <= y.supp_ind + y.supp_len - 1))) \
                or (y.supp_ind + y.supp_len - 1 > 2**j - 1 and ((0 >= y.supp_ind
                                                                 or 0 <= (y.supp_ind + y.supp_len - 1) % 2**j)
                                                                and (2**j - 1 >= y.supp_ind
                                                                     or 2**j - 1
                                                                     <= (y.supp_ind + y.supp_len - 1) % 2**j))):
            if y.supp_len != 2**j:
                # If the support of y^(j) is not the whole interval 0,...,2^j-1, we can uniquely determine the support
                # structure.
                y_0 = y.vec[y.supp_ind: 2**j]
                y_1 = y.vec[2**j + y.supp_ind: 2**(j+1)]
                # Find the significant large entries of y.
                T_0 = np.nonzero(y_0 > epsilon)[0] + y.supp_ind
                T_1 = np.nonzero(y_1 > epsilon)[0] + 2**j + y.supp_ind
                # The index sets T_0 and T_1 cannot both be empty.
                if T_0.shape[0] == 0:
                    # If T_0 is empty, then y has only a one-block support and 0, 2^J-1 are contained in it.
                    y.update_supp('one', T_1[0], 2 * (2**(j+1) - T_1[0]))
                    return y
                elif T_1.shape[0] == 0:
                    # If T_1 is empty, then y has only a one-block support and 2^(J-1)-1, 2^(J-1) are contained in it.
                    y.update_supp('one', T_0[0], 2 * (2**j - T_0[0]))
                    return y
                else:
                    # If neither set is empty, then y has a two-block support with two blocks that might have different
                    # lengths. This can only happen if 0, 2^(J-1)-1 are contained in the support of x.
                    y.update_supp('two_diff', [T_0[0], T_1[0]], [2 * (2**j - T_0[0]), 2 * (2**J - T_1[0])])
                    return y
            else:
                # If the support of y^(j) is the whole interval, the support cannot always be uniquely recovered, as
                # there are no unique first support index and block length of y^(j) that one can use for y^(j+1).
                y_0 = y.vec[0: round(y.supp_len / 2)]
                y_1 = y.vec[round(y.supp_len / 2): 2**j]
                # Find the significantly large entries of the first half of y.
                T_0 = np.nonzero(y_0 > epsilon)[0]
                T_1 = np.nonzero(y_1 > epsilon)[0] + round(y.supp_len / 2)
                # The index sets T_0 and T_1 cannot both be empty.
                if T_0.shape[0] == 0:
                    # If T_0 is empty, then y has only a one-support and 2^(J-1)-1, 2^(J-1) are contained in it.
                    y.update_supp('one', T_1[0], 2 * (2**(J-1) - T_1[0]))
                    return y
                elif T_1.shape[0] == 0:
                    # If T_1 is empty, then y has only a one-support and 0, 2^J-1 are contained in it.
                    y.update_supp('one', 2**J - 1 - T_0[-1], 2 * (T_0[-1] + 1))
                    return y
                else:
                    # If neither set is empty, there are three possibilitis for the support of y, and we do not know the
                    # correct one a priori. Either
                    # - S^(J) = {T_0[0],...,2^(J-1)-1,2^(J-1),...,2^J-1-T_0[0]},
                    # - S^(J) = {2^J-1-T_1[-1],...,2^J-1,0,...,T_1[-1],
                    # - y^(J) has a two-block support with two blocks of possibly different lengths, where neither the
                    # position of the blocks nor their length is known.
                    # The first and second case can only happen if the first support index of x is 0 or the last support
                    # index of x is 2^(J-1)-1, respectively.
                    print('Support is either {},...,{},{},...,{} or {},...,{},0,...,{} or a two-block support with '
                          'unknown block lengths and positions.'.format(T_0[0], 2**(J - 1) - 1, 2**(J - 1),
                                                                        2**J - 1 - T_0[0], 2**J - 1 - T_1[-1], 2**J - 1,
                                                                        T_1[-1]))
                    y.update_supp('two_diff', [T_0[0], 2 ** J - 1 - T_1[-1]],
                                  [2 * (2 ** (J - 1) - T_0[0]), 2 * (T_1[-1] + 1)])
                    return y


def find_supp_2(y, j, J, epsilon):
    """
    Finds the support structure of y = y^(j+1) if y^(j) has a one-block support of length m <= 2^(j-1).
    :param y: vector whose support we want to determine
    :param j: y has length 2^(j+1)
    :param J: full vector has length 2^J
    :param epsilon: noise threshold
    :return supp_struct: support structure of y
             'one' if y has a one-block support
             'two' if y has a two-block support
             'two_diff' if y has a two-block support with possibly different lengths, only possible for j = J-1
    :return mu_1: first support index of the first block of y^(j+1)
    :return m_1: length of the first block of y^(j+1)
    :return mu_2: first support index of the second block of y^(j+1)
    :return m_2: length of the second block of y^(j+1)
    """
    if j < J-1:
        # If 2^(j-1)-1, 2^(j-1) are contained in S^(j), then y^(j+1) has a two-block support.
        if (y.supp_ind + y.supp_len - 1 <= 2**j - 1 and ((2**(j-1) - 1 >= y.supp_ind)
                                                         and (2**(j-1) - 1 <= y.supp_ind + y.supp_len - 1)
                                                         and (2**(j-1) >= y.supp_ind)
           and (2**(j-1) <= y.supp_ind + y.supp_len - 1))) or (y.supp_ind + y.supp_len - 1 > 2**j - 1
                                                               and ((2**(j-1) - 1 >= y.supp_ind
                                                                     or 2**(j-1) - 1
                                                                     <= (y.supp_ind + y.supp_len - 1) % 2**j)
                                                                    and (2**(j-1) >= y.supp_ind
                                                                         or 2**(j-1)
                                                                         <= (y.supp_ind + y.supp_len - 1) % 2**j))):
            y_0 = y.vec[y.supp_ind: 2**j - y.supp_ind]
            # Find the first support index of the first block.
            T = np.nonzero(y_0 > epsilon)[0] + y.supp_ind
            y.update_supp('two', T[0], T[-1] - T[0] + 1)
            return y

        # If 0, 2^j-1 are contained in S^(j), then y^(j+1) has a one-block support.
        elif (y.supp_ind + y.supp_len - 1 <= 2**j - 1 and ((0 >= y.supp_ind) and (0 <= y.supp_ind + y.supp_len - 1)
                                                           and (2**j - 1 >= y.supp_ind)
                                                           and (2**j - 1 <= y.supp_ind + y.supp_len - 1))) \
                or (y.supp_ind + y.supp_ind - 1 > 2**j - 1 and ((0 >= y.supp_ind or 0
                                                                 <= (y.supp_ind + y.supp_len - 1) % 2**j)
                                                                and (2**j - 1 >= y.supp_ind
                                                                     or 2**j - 1
                                                                     <= (y.supp_ind + y.supp_len - 1) % 2**j))):
            # Find the first support index of the block.
            e_0 = y.vec[y.supp_ind: y.supp_ind + y.supp_len].sum()
            e_1 = y.vec[(np.arange(y.supp_ind, y.supp_ind + y.supp_len) + 2**j + 1) % 2**(j+1)].sum()
            if e_0 > e_1:
                mu_1 = y.supp_ind
            elif e_0 < e_1:
                mu_1 = y.supp_ind + 2**j
            else:
                mu_1 = 0
            y.update_supp('one', mu_1, y.supp_len)
            return y

    else:
        # If j=J-1, we have the same cases as before, but it might also happen that y has a two-block support with two
        # blocks of different lengths.
        # If 2^(j-1)-1, 2^(j-1) are contained in S^(j), then y has a two-block support.
        if (y.supp_ind + y.supp_len - 1 <= 2**j - 1 and ((2**(j - 1) - 1 >= y.supp_ind)
                                                         and (2**(j - 1) - 1 <= y.supp_ind + y.supp_len - 1)
                                                         and (2**(j - 1) >= y.supp_ind)
                                                         and (2**(j - 1) <= y.supp_ind + y.supp_len - 1))) \
                or (y.supp_ind + y.supp_len - 1 > 2**j - 1 and ((2**(j - 1) - 1 >= y.supp_ind
                                                                 or 2**(j - 1) - 1
                                                                 <= (y.supp_ind + y.supp_len - 1) % 2**j)
                                                                and (2**(j - 1) >= y.supp_ind
                                                                     or 2**(j - 1)
                                                                     <= (y.supp_ind + y.supp_len - 1) % 2**j))):
            y_0 = y.vec[y.supp_ind: 2**j - y.supp_ind]
            # Find the first support index of the first block.
            T = np.nonzero(y_0 > epsilon)[0] + y.supp_ind
            y.update_supp('two', T[0], T[-1] - T[0] + 1)
            return y
        # If 0, 2^j-1 are contained in S^(j), then y has a two-block support with possibly different block lengths, and
        # one of the blocks might be empty.
        elif (y.supp_ind + y.supp_len - 1 <= 2**j - 1 and ((0 >= y.supp_ind) and (0 <= y.supp_ind + y.supp_len - 1)
                                                           and (2**j - 1 >= y.supp_ind)
                                                           and (2**j - 1 <= y.supp_ind + y.supp_len - 1))) \
                or (y.supp_ind + y.supp_len - 1 > 2**j - 1 and ((0 >= y.supp_ind
                                                                 or 0 <= (y.supp_ind + y.supp_len - 1) % 2**j)
                                                                and (2**j - 1 >= y.supp_ind
                                                                     or 2**j - 1
                                                                     <= (y.supp_ind + y.supp_len - 1) % 2**j))):
            y_0 = y.vec[y.supp_ind: 2**j]
            y_1 = y.vec[2**j + y.supp_ind:]
            # Find the significant entries of y.
            T_0 = np.nonzero(y_0 > epsilon)[0] + y.supp_ind
            T_1 = np.nonzero(y_1 > epsilon)[0] + 2**j + y.supp_ind
            # Not both index sets T_0 and T_1 can be empty.
            if T_0.shape[0] == 0:
                # If T_0 is empty, then y has only a one-block support, and 0, 2^J-1 are contained in it.
                y.update_supp('one', T_1[0], 2 * (2**(j + 1) - T_1[0]))
                return y
            elif T_1.shape[0] == 0:
                # If T_1 is empty, then y has only a one-block support, and 2^(J-1)-1, 2^(J-1) are contained in it.
                y.update_supp('one', T_0[0], 2 * (2**j - T_0[0]))
                return y
            else:
                # If neither set is empty, then y has a two-block support with two blocks that might have different
                # lengths. This can only happen if 0, 2^(J-1)-1 are contained in the support of x.
                y.update_supp('two_diff', [T_0[0], T_1[0]], [2 * (2**j - T_0[0]), 2 * (2**J - T_1[0])])
                return y


def find_supp_3(y, j, l):
    """
    Finds the support structure of y^(j+1)=y if y^(j) has a two-block support.
    :param y: vector whose support we want to determine
    :param j: y^(j+1) has length 2^(j+1)
    :param l: parameter depending on the support structure of y^(j+1)
    :return supp_struct: support structure of y
         'one' if y has a one-block support
         'two' if y has a two-block support
         'two_diff' if y has a two-block support with possibly different lengths, only possible for j = J-1
    :return mu_1: first support index of the first block of y^(j+1)
    :return m_1: length of the first block of y^(j+1)
    """

    # The first support index of the first block only depends on the previously computed values l, mu, m and j. y^(j)
    # and y^(j+1) both have a two-block support with the same block length.
    if l == y.supp_ind:
        mu_1 = y.supp_ind
    elif l == 2**j + y.supp_ind:
        mu_1 = 2**j - y.supp_len - y.supp_ind
    else:
        mu_1 = 0
    y.update_supp('two', mu_1, y.supp_len)
    return y
