import numpy as np
import math
import random
from time import time
from scipy.fftpack import fft, ifft
import holoviews as hv
import DFT_twoblock_mod
hv.extension('bokeh')
renderer = hv.renderer('bokeh')

# Input for recreating the figures in "Real Sparse Fast DCT for Vectors with Short Support" by Sina Bittens and Gerlind
# Plonka.
N = 2**20
tests = 100
epsilon = 1e-4
variant = 'exact'
# Input exact data
m = np.array([10, 50, 100, 1000, 10000, 50000, 100000])
# Input noisy data
# m = np.array([100, 1000])


def test_dft_twoblock(tests, m, N, epsilon, variant='exact'):
    """
    Computes the average runtime and reconstruction error of DCT_real, DCT_shortsupp and idct.
    :param tests: Number of tries
    :param m: support length
    :param N: vector length
    :param epsilon: noise threshold
    :param variant: variant=='exact': exact data; variant=='noisy': noisy data (uniform noise)
    :return:
    """

    m = m.ravel()
    mm = m.shape[0]
    # N needs to be a power of 2, and m, M need to be arrays of the same shape with m <= M componentwise.
    if abs(math.log(N, 2) - round(math.log(N, 2))) < np.spacing(1):
        J = math.log(N, 2) + 1
        # Initialize the matrix for the average runtimes.
        time_av = np.zeros((mm, 4))
        time_av[:, 0] = m
        # Define the SNR values for noisy signals.
        SNR = np.arange(0, 60, 10)
        l_SNR = SNR.shape[0]
        # Define the cutoff thresholds for noisy data, optimized for N=2^20 and m=100 or m=1000.
        epsilon_noise = np.array([1.7, 1.2, 0.4, 0.19, 0.05, 0.02])

        # Initialize the matrix of reconstruction errors for exact data.
        error_ex = np.zeros((mm, 2))
        plot_list = []
        for j in range(mm):
            # Initialize the matrix for the average runtimes for noisy data.
            time_noise = np.zeros((l_SNR, 3))
            time_noise[:, 0] = SNR
            # Initialize the matrix for the average noisy reconstruction errors.
            error = np.zeros((l_SNR, 3))
            error[:, 0] = SNR
            # Initialize the matrix for the maximal noisy reconstruction errors.
            max_error = np.zeros((l_SNR, 3))
            max_error[:, 0] = SNR
            # Initialize the matrix for the rate of correct recovery.
            correct_recov = np.zeros((l_SNR, 3))
            correct_recov[:, 0] = SNR

            for t in range(tests):
                if t % 10 == 1:
                    print('{}st test'.format(t))
                # For each test create a random first support index mu for the support block in the first half of y,
                # which has to be a vector of length 2N with reflected support of block length m[j].
                mu = random.randint(0, N - 1)
                y = np.zeros(2 * N)
                # Generate a random number between 0 and floor((m[j] - 2)/2) indices inside the support of the first
                # block, excluding the first and last support index, whose corresponding entries will be set to zero.
                supp_zero = np.array(random.sample(range(mu + 1, mu + m[j] - 1),
                                                   random.randint(0, math.floor((m[j] - 2) / 2))))
                supp_zero = supp_zero % N
                # Generate uniformly distributed entries between 0 and 10. The entries corresponding to the first and
                # last support index have to be significantly large, i.e., > epsilon and > epsilon_noise.
                if variant == 'exact':
                    y[np.array([mu, (mu + m[j] - 1) % N])] = epsilon + (10 - epsilon) * np.random.random(2)
                elif variant == 'noisy':
                    y[np.array([mu, (mu + m[j] - 1) % N])] \
                        = epsilon_noise[j] + (10 - epsilon_noise[-1]) * np.random.random(2)
                y[np.array([k for k in (np.arange(mu + 1, mu + m[j] - 1)) % N if k not in supp_zero])] \
                    = np.random.random(m[j] - len(supp_zero) - 2) * 10

                # The second half of y is already determined by the first one, as y is symmetric.
                y[N:] = np.flip(y[0: N], 0)
                ####
                y = np.round(y)
                # y[0:N] = [0, 3, 8, 0, 0,0,0,0, 0,0,0,0, 0,0,6,0]
                # y[N:] = np.flip(y[0: N], 0)

                # Generate the input data.
                yhat = fft(y)

                # Standard IFFT Algorithm
                # Measure the runtime of ifft.
                start = time()
                s_ifft = ifft(yhat)
                time_av[j, 2] += time() - start

                # Numerical Experiments for Exact Input Data
                if variant == 'exact':
                    # dft_twoblock
                    # Measure the runtime of dft_twoblock.
                    start = time()
                    s_twoblock = DFT_twoblock_mod.dft_twoblock(yhat, epsilon).vec
                    time_av[j, 1] += time() - start

                    # Compute the reconstruction errors for exact data for the two methods, given by the 2-norm of the
                    # difference between the original vector y and the returned vectors s_twoblock and s_ifft.
                    error_ex[j, 0] += np.linalg.norm(s_twoblock - y) / (2 * N)
                    error_ex[j, 1] += np.linalg.norm(s_ifft - y) / (2 * N)

                # Numerical Experiments for Noisy Input Data
                elif variant == 'noisy':
                    for n in range(l_SNR):
                        # Generate uniformly distributed noise for the input vector yhat.
                        noise = np.random.random(2 * N) - 0.5 + 1j * np.random.random(2 * N) - 0.5
                        # Scale the noise according to the current SNR value SNR[n].
                        noise = np.linalg.norm(yhat) * 10 ** (-SNR[n] / 20) / np.linalg.norm(noise) * noise
                        # Create noisy input data by addition.
                        yhat_noise = yhat + noise

                        # Standard IFFT Algorithm
                        # Measure the runtime of ifft for noisy input data.
                        start = time()
                        s_ifft_noise = ifft(yhat_noise)
                        time_noise[n, 2] += time() - start

                        # dft_twoblock
                        # Measure the runtime of dft_twoblock for noisy input data.
                        start = time()
                        s_twoblock_noise = DFT_twoblock_mod.dft_twoblock(yhat_noise, epsilon_noise[n])
                        time_noise[n, 1] += time() - start

                        # Check whether the support of y was found correctly.
                        # First examine whether the support of y is contained in the returned support of
                        # s_twoblock_noise.
                        if s_twoblock_noise.structure == 'two_diff':
                            s_twoblock_noise.supp_ind = s_twoblock_noise.supp_ind[0]
                            s_twoblock_noise.supp_len = s_twoblock_noise.supp_len[0]
                        if s_twoblock_noise.supp_ind + s_twoblock_noise.supp_len <= 2 ** J - 1:
                            if (mu >= s_twoblock_noise.supp_ind
                                    and mu + m[j] - 1 <= s_twoblock_noise.supp_ind + s_twoblock_noise.supp_len - 1):
                                correct_recov[n, 1] += 1
                                # We also check whether the returned support length of s.twoblock_noise is at most three
                                # times the support length of y
                                if s_twoblock_noise.supp_len <= 3 * m[j] and s_twoblock_noise.supp_len >= m[j]:
                                    correct_recov[n, 2] += 1
                        else:
                            # If 0, 2^J-1 are contained in the support of s_twoblock_noise, the block has to be
                            # considered modulo 2^J.
                            if ((mu >= s_twoblock_noise.supp_ind
                                 or mu <= s_twoblock_noise.supp_ind + s_twoblock_noise.supp_len - 1 - 2 ** J)
                                    and (mu + m[j] - 1 >= s_twoblock_noise.supp_ind
                                         or mu + m[j] - 1
                                         <= s_twoblock_noise.supp_ind + s_twoblock_noise.supp_len - 1 - 2 ** J)):
                                correct_recov[n, 1] += 1
                                if s_twoblock_noise.supp_len <= 3 * m[j] and s_twoblock_noise.supp_len >= m[j]:
                                    correct_recov[n, 2] += 1
                        # Compute the reconstruction errors for noisy data for the two methods.
                        norm_1 = np.linalg.norm(s_twoblock_noise.vec - y) / (2 * N)
                        norm_2 = np.linalg.norm(s_ifft_noise - y) / (2 * N)
                        error[n, 1] += norm_1
                        error[n, 2] += norm_2
                        # Compute the maximal reconstruction errors for noisy data for the two methods.
                        if norm_1 > max_error[n, 1]:
                            max_error[n, 1] = norm_1
                        if norm_2 > max_error[n, 2]:
                            max_error[n, 2] = norm_2

            if variant == 'noisy':
                time_noise[:, 1:] = time_noise[:, 1:] / tests
                error[:, 1:] = error[:, 1:] / tests
                print(error)
                correct_recov[:, 1:] = correct_recov[:, 1:] / tests

                # Create Plots
                # Average Reconstruction Errors
                lines_twoblock = hv.Curve((error[:, 0], error[:, 1]), label="dft_twoblock").options(
                    color='blue')
                points_twoblock = hv.Points((error[:, 0], error[:, 1])).options(color='blue', size=8, marker='v')
                lines_ifft = hv.Curve((error[:, 0], error[:, 2]), label="ifft").options(color='lime')
                points_ifft = hv.Points((error[:, 0], error[:, 2])).options(color='lime', size=8, marker='d')
                twoblock = (lines_twoblock * points_twoblock)
                ifft_plot = (lines_ifft * points_ifft)

                plot_av_err = (twoblock * ifft_plot) \
                    .options(logy=True, xlabel='SNR', ylabel='Average Reconstruction Error', legend_position='right',
                             width=750, height=300, show_grid=True) \
                    .relabel('Average Reconstruction Error, m = {}'.format(m[j]))
                renderer.save(plot_av_err, 'average_error_DFT_{}'.format(m[j]))
                plot_list.append(plot_av_err)

                # Maximal Reconstruction Errors
                lines_twoblock = hv.Curve((max_error[:, 0], max_error[:, 1]), label="dft_twoblock") \
                    .options(color='blue')
                points_twoblock = hv.Points((max_error[:, 0], max_error[:, 1])).options(color='blue', size=8,
                                                                                        marker='v')
                lines_ifft = hv.Curve((max_error[:, 0], max_error[:, 2]), label="ifft").options(color='lime')
                points_ifft = hv.Points((max_error[:, 0], max_error[:, 2])).options(color='lime', size=8, marker='d')
                twoblock = (lines_twoblock * points_twoblock)
                ifft_plot = (lines_ifft * points_ifft)

                plot_max_err = (twoblock * ifft_plot) \
                    .options(logy=True, xlabel='SNR', ylabel='Maximal Reconstruction Error', legend_position='right',
                             width=750, height=300, show_grid=True) \
                    .relabel('Maximal Reconstruction Error, m = {}'.format(m[j]))
                renderer.save(plot_max_err, 'maximal_error_DFT_{}'.format(m[j]))
                plot_list.append(plot_max_err)

        # Compute the average runtimes and the average reconstruction errors.
        time_av[:, 1:] = time_av[:, 1:] / tests
        print(time_av)
        error_ex = error_ex / tests
        print('error_ex={}'.format(error_ex))

        # Create Plots
        if variant == 'exact':
            # Runtime
            lines_twoblock = hv.Curve((time_av[:, 0], time_av[:, 1]), label="dft_twoblock") \
                .options(color='blue')
            points_twoblock = hv.Points((time_av[:, 0], time_av[:, 1])).options(color='blue', size=8,
                                                                                marker='v')
            lines_ifft = hv.Curve((time_av[:, 0], time_av[:, 2]), label="ifft").options(color='lime')
            points_ifft = hv.Points((time_av[:, 0], time_av[:, 2])).options(color='lime', size=8, marker='d')
            twoblock = (lines_twoblock * points_twoblock)
            ifft_plot = (lines_ifft * points_ifft)

            plot_runtime = (twoblock * ifft_plot) \
                .options(logx=True, logy=True, xlabel='Support Length m', ylabel='Runtime [s]',
                         legend_position='right',
                         width=750, height=300, show_grid=True).relabel('Average Runtime')
            renderer.save(plot_runtime, 'runtime_DFT')

        elif variant == 'noisy':
            plot_errors = hv.Layout(plot_list).cols(2)
            renderer.save(plot_errors, 'errors_DFT')

    else:
        print('One of the parameters has been set incorrectly.')


test_dft_twoblock(tests, m, N, epsilon, variant)
